
#
#   YaLafi module for LaTeX package amsthm
#

from yalafi.defs import InitModule, Environ, SpaceToken, TextToken

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = r"""

        \newcommand{\qedhere}{}
        \newcommand{\theoremstyle}[1]{}
        \newcommand{\newtheoremstyle}[9]{}

    """

    macros_python = []

    environments = [

        Environ(parms, 'proof', args='O', repl=h_proof),

    ]

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)

def h_proof(parser, buf, mac, args, delim, pos):
    if args[0]:
        ret = args[0]
    else:
        ret = [TextToken(pos, parser.parms.lang_context.proof_name,
                        pos_fix=True)]
    return ret + [TextToken(ret[-1].pos, '.', pos_fix=True),
                        SpaceToken(ret[-1].pos, '\n', pos_fix=True)]





#
#   YaLafi module for LaTeX package xcolor
#

from yalafi.defs import Macro, InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = r"""

        \newcommand{\color}[2][]{}
        \newcommand{\colorbox}[3][]{#3}
        \newcommand{\definecolor}[4][]{}
        \newcommand{\textcolor}[3][]{#3}

    """

    macros_python = [

        Macro(parms, '\\fcolorbox', args='OAOAA', repl='#5'),

    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)




#
#   YaLafi: Yet another LaTeX filter
#   Copyright (C) 2020 Matthias Baumann
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from . import defs, parameters
import copy
import re
import sys


#   print error message to stderr,
#   return token sequence suitable to be inserted into filter output
#
def latex_error(err, pos, latex, parms):
    lin = latex.count('\n', 0, pos) + 1
    nl = latex.rfind('\n', 0, pos) + 1
    col = pos - nl + 1
    sys.stderr.write('*** LaTeX error: line ' + str(lin)
                        + ', column ' + str(col) + ':\n*** ' + err + '\n')
    sys.stderr.flush()
    mark = ' ' + parms.mark_latex_error + ' '
    if parms.mark_latex_error_verbose:
        mark += '(' + err + ') '
    mx = min(len(mark), len(latex) - pos)
    out = [defs.TextToken(pos, mark[:mx], pos_fix=True)]
    if mx < len(mark):
        out.append(defs.TextToken(pos + mx -1, mark[mx:], pos_fix=True))
    return out

def fatal(err):
    sys.stderr.write('*** ' + sys.argv[0] + ': internal error:\n*** '
                        + err + '\n')
    sys.exit(1)

def warning(err):
    sys.stderr.write('*** ' + sys.argv[0] + ': warning:\n*** '
                        + err + '\n')
    sys.stderr.flush()

def get_txt_pos(toks):
    txt = ''
    pos = []
    for t in toks:
        txt += t.txt
        if t.pos_fix:
            pos += [t.pos] * len(t.txt)
        else:
            pos += list(range(t.pos, t.pos + len(t.txt)))
    return txt, pos

def replace_phrases(txt, pos, lines):
    for lin in lines:
        i = lin.find('#')
        if i >= 0:
            lin = lin[:i]
        lin = lin.split()

        t = s = ''
        for i in range(len(lin)):
            if lin[i] == '&':
                break
            t += s + re.escape(lin[i])
                    # protect e.g. '.' and '$'
            s = r'(?:[ \t]*\n[ \t]*|[ \t]+)'
                    # at least one space character, but stay in paragraph
        if not t:
            continue
        if t[0].isalpha():
            t = r'\b' + t       # require word boundary
        if t[-1].isalpha():
            t = t + r'\b'

        r = ' '.join(lin[i+1:])
        txt, pos = substitute(txt, pos, t, r)
    return txt, pos

def substitute(i_txt, i_pos, expr, repl):
    o_txt = ''
    o_pos = []
    r_len = len(repl)
    last = 0
    for m in re.finditer(expr, i_txt):
        m_len = len(m.group(0))
        if not m_len:
            continue
        cur = m.start(0)
        o_txt += i_txt[last:cur] + repl
        o_pos += i_pos[last:cur]
        if r_len <= m_len:
            o_pos += i_pos[cur:cur+r_len]
        else:
            o_pos += (i_pos[cur:cur+m_len]
                        + [i_pos[cur+m_len-1]] * (r_len - m_len))
        last = m.end(0)
    return o_txt + i_txt[last:], o_pos + i_pos[last:]

#   get handler for importing a module:
#   - replace non-alphanumeric characters (except '.') in LaTeX package name
#     by '_'
#   - if module name starts with '.': remove that dot
#   - else: prepend given prefix
#   return handler, dummy on error
#
def get_module_handler(name, prefix):
    name = ''.join((c if c.isalnum() or c == '.' else '_') for c in name)
    if name.startswith('.'):
        mod = name[1:]
    else:
        mod = prefix + '.' + name
    try:
        exec('import ' + mod)
        return (eval(mod + '.require_packages'),
                        eval(mod + '.init_module'))
    except:
        warning('could not load module ' + repr(mod))
        return [], lambda p, o, n: defs.InitModule()

#   filter out tokens and set their position numbers
#
def filter_set_toks(toks, pos, tok_typ):
    def f(t):
        t = copy.copy(t)
        t.pos = pos
        return t
    return [f(t) for t in toks if tok_typ is None or type(t) is tok_typ]


class LanguageSection:
    def __init__(self, lang, back, brk, txt, pos):
        self.lang = lang
        self.back = back        # e.g., started by end of \foreignlanguage
        self.brk = brk          # e.g., started by \selectlanguage
        self.txt = txt
        self.pos = pos

#   transform a token list into text strings and position lists
#   - seperate into different languages
#   - heuristic: see comments below marked with HEURISTIC
#
def get_txt_pos_ml(toks, main_lang, parms):

    lang_stack = [main_lang]
    switch_back = False
    switch_brk = False

    #   split into sections separated by language switches
    #
    sections = []
    cur_sec = []
    for t in toks:
        if type(t) is not defs.LanguageToken:
            cur_sec.append(t)
            continue
        if t.lang == lang_stack[-1]:
            continue
        txt, pos = get_txt_pos(cur_sec)
        cur_sec = []
        if txt:
            sections.append(LanguageSection(
                        lang_stack[-1], switch_back, switch_brk, txt, pos))
        switch_back = t.back
        switch_brk = t.brk
        if t.back:
            if len(lang_stack) > 1:
                lang_stack.pop()
        else:
            if t.hard:
                lang_stack[-1] = t.lang
            else:
                lang_stack.append(t.lang)
    txt, pos = get_txt_pos(cur_sec)
    if txt:
        sections.append(LanguageSection(
                        lang_stack[-1], switch_back, switch_brk, txt, pos))

    #   try to combine sections, if only interrupted by short change to
    #   another language
    #
    out = []
    while sections:
        if (len(sections) > 1
                    # HEURISTIC to connect sections:
                    # ==============================
                    # - next section does not mandatorily break the text flow
                    #   (e.g., caused by \selectlanguage)
                and not sections[1].brk
                    # - next section must not start with a switch_back
                    #   (e.g., section not after \foreignlanguage call)
                and not sections[1].back
                    # - following section (if any) has to be same language
                and (len(sections) < 3
                            or sections[0].lang == sections[2].lang)
                    # - next section has to be "suitable" to be replaced
                and ml_check_lang_section(sections[1], parms)):

            # we have a short inclusion with another language
            incl = sections.pop(1)
            out.append(incl)
            ml_append_placeholder(sections[0], incl, parms)
            if len(sections) > 1:
                # append following section with same language
                sections[0].txt += sections[1].txt
                sections[0].pos += sections[1].pos
                sections.pop(1)
        else:
            out.append(sections.pop(0))

    #   combine all sections of same language into one list
    #
    ret = {}
    for sec in out:
        if sec.lang in ret:
            ret[sec.lang].append([sec.txt, sec.pos])
        else:
            ret[sec.lang] = [[sec.txt, sec.pos]]
    return ret

#   append placeholder for "short" language switch to current section
#
def ml_append_placeholder(sec, incl, parms):
    if not incl.txt.strip():
        # inclusion is empty or only contains space
        sec.txt += incl.txt
        sec.pos += incl.pos
        return

    lang = parms.check_parser_lang(sec.lang)
    repl = parms.parser_lang_settings[lang].lang_change_repl
    repl[:] = repl [1:] + repl[:1]
    txt = repl[0]

    start = next((n for n in range(len(incl.txt))
                            if not incl.txt[n].isspace()), 0)
    pos = [incl.pos[start]] * len(repl[0])

    # see issue #117
    if incl.txt.strip():
        if incl.txt[0].isspace():
            txt = incl.txt[0] + txt
            pos = [incl.pos[0]] + pos
        if incl.txt[-1].isspace():
            txt += incl.txt[-1]
            pos.append(incl.pos[-1])

    sec.txt += txt
    sec.pos += pos

#   heuristic to determine whether a (short) language section
#   can be substituted with a placeholder, so that previous and
#   subsequent sections can be glued together
#
def ml_check_lang_section(sec, parms):
    # accept, if less than 4 words
    return len(sec.txt.split()) <= parms.ml_continue_thresh





#
#   YaLafi module for LaTeX package biblatex
#
#   very simple approximation
#   - citation text is fixed --> variable cite_text
#   - we always add [] brackets (even for \cite and \footcite)
#

from yalafi import defs
from yalafi.defs import Macro, InitModule

require_packages = []

cite_text = '0'

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = ''

    macros_python = [

        Macro(parms, '\\addbibresource', args='OA', repl=''),
        Macro(parms, '\\cite', args='*OOA', repl=h_cite),
        Macro(parms, '\\Cite', args='*OOA', repl=h_cite),
        Macro(parms, '\\footcite', args='*OOA', repl=h_footcite),
        Macro(parms, '\\footcitetext', args='*OOA', repl=h_footcite),
        Macro(parms, '\\parencite', args='*OOA', repl=h_cite),
        Macro(parms, '\\Parencite', args='*OOA', repl=h_cite),
        Macro(parms, '\\printbibliography', args='O', repl=''),

    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)


def h_cite(parser, buf, mac, args, delim, pos):
    opt1 = args[1]
    opt2 = args[2]
    if len(opt1) == 1 and type(opt1[0]) is defs.VoidToken:
        # only [] given
        opt1 = []
    if opt2:
        pre = opt1
        if len(opt2) == 1 and type(opt2[0]) is defs.VoidToken:
            # only [] given
            opt2 = []
        post = opt2
    else:
        pre = []
        post = opt1
        
    out = [defs.TextToken(pos, '[', pos_fix=True)]
    if pre:
        out += pre
        out.append(defs.SpaceToken(out[-1].pos, ' ', pos_fix=True))
    out.append(defs.TextToken(out[-1].pos, cite_text, pos_fix=True))
    if post:
        out += [defs.TextToken(out[-1].pos, ',', pos_fix=True),
                    defs.SpaceToken(out[-1].pos, ' ', pos_fix=True)]
        out += post
    out += [defs.TextToken(out[-1].pos, ']', pos_fix=True),
                defs.ActionToken(out[-1].pos)]
    return out

def h_footcite(parser, buf, mac, args, delim, pos):
    out = [defs.MacroToken(pos, '\\footnote'),
                defs.SpecialToken(pos, '{')]
    out += h_cite(parser, buf, mac, args, delim, pos)
    out += [defs.TextToken(out[-1].pos, '.', pos_fix=True),
                    defs.SpecialToken(out[-1].pos, '}'),
                    defs.ActionToken(out[-1].pos)]
    return out





#
#   YaLafi module for LaTeX package babel
#

#   do the macros / environments always break the text flow?
#
foreignlang_break = False
selectlang_break = True
otherlang_break = False

#   map between language codes for babel and LanguageTool
#   - sorted according to xx-XX code
#
language_map = {
#   'ar'
#   'ast-ES'
    'belarusian': 'be-BY',
    'breton': 'br-FR',
    'catalan': 'ca-ES',
#   'ca-ES-valencia'
    'danish': 'da-DK',
    'german': 'de-DE',
    'ngerman': 'de-DE',
    'german-at': 'de-AT',
    'german-ch': 'de-CH',
    'greek': 'el-GR',
    'english': 'en-GB',
    'english-au': 'en-AU',
    'english-ca': 'en-CA',
    'british': 'en-GB',
    'english-gb': 'en-GB',
    'english-nz': 'en-NZ',
    'american': 'en-US',
    'english-us': 'en-US',
#   'en-ZA',
    'esperanto': 'eo',
    'spanish': 'es',
#   'fa',
    'french': 'fr',
#   'ga-IE'
    'galician': 'gl-ES',
    'italian': 'it',
    'japanese': 'ja-JP',
#   'km-KH'
    'dutch': 'nl',
#   'nl-BE'
    'polish': 'pl-PL',
    'portuguese': 'pt-PT',
#   'pt-AO'
    'portuguese-br': 'pt-BR',
#   'pt-MZ'
    'romanian': 'ro-RO',
    'russian': 'ru-RU',
    'slovak': 'sk-SK',
    'slovenian': 'sl-SI',
    'swedish': 'sv',
    'tamil': 'ta-IN',
#   'tl-PH'
    'ukrainian': 'uk-UA',
    'chinese': 'zh-CN',
}

from yalafi.defs import InitModule, Macro, LanguageToken, Environ, MacroToken

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = ''

    macros_python = [

        # the following is for \end{otherlanguage}
        Macro(parms, '\\babel@skip@space', args='', repl=''),
        Macro(parms, '\\foreignlanguage', args='OAA', repl=h_foreignlanguage),
        Macro(parms, '\\selectlanguage', args='A', repl=h_selectlanguage),

    ]

    environments = [

        Environ(parms, 'otherlanguage', args='A', repl=h_begin_otherlang,
                                add_pars=False, end_func=h_end_otherlang),
        Environ(parms, 'otherlanguage*', args='A', repl=h_begin_otherlang,
                                add_pars=False, end_func=h_end_otherlang_star),

    ]

    inject_tokens = get_language_token(parser.global_latex_options + options)

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments, inject_tokens=inject_tokens)

def modify_language_map(babel, lt):
    language_map[babel] = lt

#   translate babel code to LanguageTool code
#
def translate_lang(lang):
    return language_map.get(lang, language_map['english'])

def h_foreignlanguage(parser, buf, mac, args, delim, pos):
    lang = translate_lang(parser.get_text_expanded(args[1]).strip())
    return ([LanguageToken(pos, lang=lang, brk=foreignlang_break)] + args[2]
                        + [LanguageToken(args[2][-1].pos, back=True)])

def h_selectlanguage(parser, buf, mac, args, delim, pos):
    lang = translate_lang(parser.get_text_expanded(args[0]).strip())
    return [LanguageToken(pos, lang=lang, hard=True, brk=selectlang_break)]

def h_begin_otherlang(parser, buf, mac, args, delim, pos):
    lang = translate_lang(parser.get_text_expanded(args[0]).strip())
    return [LanguageToken(pos, lang=lang, brk=otherlang_break)]

def h_end_otherlang(parser, buf, mac, args, delim, pos):
    return [LanguageToken(pos, back=True),
                MacroToken(pos, '\\babel@skip@space')]

def h_end_otherlang_star(parser, buf, mac, args, delim, pos):
    return [LanguageToken(pos, back=True)]

#   return a list
#   - with LanguageToken corresponding to given option
#   - or empty, if no language option found
#
def get_language_token(options):
    for opt in reversed(options):
        if opt[1] is None and opt[0] in language_map:
            return [LanguageToken(0, lang=translate_lang(opt[0]),
                                                hard=True, brk=True)]
    return []





#
#   YaLafi: Yet another LaTeX filter
#   Copyright (C) 2020 Matthias Baumann
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#
#   extract packages included by a file
#   - register as package extension module at yalafi
#   - overwrite macros \documentclass, \usepackage
#   - track calls of these macros
#

from yalafi import defs, tex2txt

require_packages = []
def init_module(parser, options, position):
    parms = parser.parms
    def add(parser, buf, mac, args, delim, pos):
        name = parser.get_text_expanded(args[1]).strip()
        if mac.name == '\\usepackage':
            for p in name.split(','):
                if p.strip():
                    packages.append(p.strip())
        else:
            documentclass[0] = name
        return []
    macros_python = [
        defs.Macro(parms, '\\documentclass', args='OA', repl=add),
        defs.Macro(parms, '\\usepackage', args='OA', repl=add),
    ]
    return defs.InitModule(macros_python=macros_python)

documentclass = ['']
packages = []
def addpacks(cmdline):
    packs = '.yalafi.shell.addpacks'
    if cmdline.packages.strip(','):
        packs = cmdline.packages.strip(',') + ',' + packs
    opts = tex2txt.Options(defs=cmdline.define, lang=cmdline.language[:2],
                                    dcls=cmdline.documentclass, pack=packs)
    f = tex2txt.myopen(cmdline.add_modules, encoding=cmdline.encoding)
    latex = f.read()
    f.close()
    tex2txt.tex2txt(latex, opts)
    return documentclass[0], packages





#
#   YaLafi module for LaTeX package glossaries
#
#   Conditions for operation
#   ------------------------
#   - We rely on the .glsdefs database file generated by package glossaries.
#     Note that the .glsdefs file is only generated, if the definitions with
#     \newacronym, \newglossaryentry etc. are not placed in the LaTeX preamble,
#     but inside of \begin{document} ... \end{document}. Be aware that you
#     might need to call twice (latex + makeglossaries + latex) in order to
#     obtain a completely updated .glsdefs file.
#   - All LaTeX files have to read this database, before \gls etc. can be
#     expanded by the filter. This may be done with
#       \newcommand{\LTinput}[1]{}  % only, if not yet seen by LaTeX
#                                   % --> better placed in document preamble
#       \LTinput{main.glsdefs}
#     Here, the LaTeX file 'main.tex' invokes \makeglossaries in the preamble.
#   - Please note that the path to main.glsdefs has to be valid at the moment
#     YaLafi is called, e.g., when starting a proofreader via an editor plugin.
#     If you invoke the editor at a subdirectory of the LaTeX project, then you
#     might have to write something like \LTinput{../main.glsdefs}.
#     A good way is to use something like `vi ch-Banach/open-mapping.tex`.
#   - Call of \LTinput{file.glsdefs} also can be done in an auxiliary
#     file that contains custom YaLafi macro definitions, say custom.tex.
#     This file in turn may be included by \LTinput{custom.tex} into each
#     LaTeX source. Alternatively, it may be passed by option --defs
#     or --define to yalafi or yalafi.shell, respectively. (In this case,
#     which does not require a change of the LaTeX source, custom.tex should
#     contain \usepackage{glossaries} before \LTinput{file.glsdefs}, or
#     the package name should be passed with options --pack / --packages.)
#
#   Behaviour of macros
#   -------------------
#   - \newacronym and \newglossaryentry create plain-text output from the
#     description. We start with a capital letter, and append a dot. The
#     latter is only done, if the text does not end in one of '.!?'.
#   - For the simple \gls commands, we always use the 'text' entry, as this
#     is available for all glossary entries in .glsdefs. The same accounts for
#     all plural forms, where we use entry 'plural'.
#   - We have not yet implemented the trailing optional argument of \gls etc.
#     Currently, this would consume subsequent space, if this argument is
#     missing; compare the behaviour of \footnotemark[] and
#     \newtheorem{}[]{}[].
#

import copy
from yalafi import defs, utils
from yalafi.defs import Macro, InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = ''

    macros_python = [

        Macro(parms, '\\gls', args='OA', repl=h_gls('text', [])),
        Macro(parms, '\\glspl', args='OA', repl=h_gls('plural', [])),
        Macro(parms, '\\Gls', args='OA', repl=h_gls('text', [cap_first])),
        Macro(parms, '\\Glspl', args='OA', repl=h_gls('plural', [cap_first])),
        Macro(parms, '\\GLS', args='OA', repl=h_gls('text', [cap_all])),
        Macro(parms, '\\GLSpl', args='OA', repl=h_gls('plural', [cap_all])),

        Macro(parms, '\\glsdesc', args='OA',
                                    repl=h_gls('description', [])),
        Macro(parms, '\\Glsdesc', args='OA',
                                    repl=h_gls('description', [cap_first])),
        Macro(parms, '\\GLSdesc', args='OA',
                                    repl=h_gls('description', [cap_all])),

        Macro(parms, '\\glsdisp', args='OAA', repl='#3'),
        Macro(parms, '\\glslink', args='OAA', repl='#3'),

        Macro(parms, '\\glstext', args='OA', repl=h_gls('text', [])),
        Macro(parms, '\\Glstext', args='OA', repl=h_gls('text', [cap_first])),
        Macro(parms, '\\GLStext', args='OA', repl=h_gls('text', [cap_all])),

        Macro(parms, '\\longnewglossaryentry', args='AAA', repl=h_newacronym),
        Macro(parms, '\\newacronym', args='AAA', repl=h_newacronym),
        Macro(parms, '\\newglossaryentry', args='AA', repl=h_newglossaryentry),

        # this is for reading the .glsdefs database
        Macro(parms, '\\gls@defglossaryentry', args='AA',
                                                    repl=h_parse_glsdefs),

    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)

#   return a handler function:
#   - read glossary entry with label given in macro argument
#   - get token list for specified key
#   - apply list of modifications
#   - set all token positions to beginning of \gls macro
#
def h_gls(key, mods):
    def f(parser, buf, mac, args, delim, pos):
        toks = get_tokens(parser, args[1], key)
        if toks is None:
            return utils.latex_error('could not find label for \\gls...'
                    + ' - did you include "\\LTinput{<main file>.glsdefs}"?',
                    pos, parser.latex, parser.parms)
        for f in mods:
            toks = f(toks)
        for t in toks:
            t.pos = pos
            t.pos_fix = True
        return toks
    return f

def h_newacronym(parser, buf, mac, args, delim, pos):
    return modify_description(parser, args[2])

def h_newglossaryentry(parser, buf, mac, args, delim, pos):
    descr = parser.parse_keyvals_dict(args[1]).get('description', [])
    return modify_description(parser, descr)

the_glossary = {}

#   main task: parse 'key={...},key={...},...' in second macro argument
#
def h_parse_glsdefs(parser, buf, mac, args, delim, pos):
    label = parser.get_text_expanded(args[0])
    the_glossary[label] = parser.parse_keyvals_dict(args[1])
    return []

#   get token list for a key, database label given as token list
#
def get_tokens(parser, label, key):
    label = parser.get_text_expanded(label)
    if label not in the_glossary or key not in the_glossary[label]:
        return None
    return the_glossary[label][key]

#   capitalise first letter
#
def cap_first(toks):
    i = next((n for n in range(len(toks))
                    if type(toks[n]) is defs.TextToken), -1)
    if i >= 0:
        # NB: keep original token list
        toks = toks.copy()
        toks[i] = copy.copy(toks[i])
        toks[i].txt = toks[i].txt[0].upper()
    return toks

#   capitalise all letters
#
def cap_all(toks):
    def f(t):
        if type(t) is defs.TextToken:
            t = copy.copy(t)
            t.txt = t.txt.upper()
        return t
    return [f(t) for t in toks.copy()]

#   modify description of glossary entry:
#   - capitalise first letter
#   - append dot
#
def modify_description(parser, toks):
    toks = cap_first(toks)
    txt = parser.get_text_expanded(toks)
    if txt and txt[-1] not in ('.', '!', '?'):
        toks.append(defs.TextToken(toks[-1].pos, '.', pos_fix=True))
    return toks





#
#   YaLafi module for LaTeX package listings
#

from yalafi.defs import InitModule, Environ

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = r"""

        \newcommand{\lstinputlisting}[2][]{\par}
        \newcommand{\lstset}[1]{}

    """

    macros_python = []

    environments = [

        Environ(parms, 'lstlisting', remove=True),

    ]

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)





#
#   YaLafi module for LaTeX package unicode-math
#

from yalafi.defs import Macro, InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    math_operators = [
        '≤', '≥', '→', '∩', '∪', '⇒', '⇐',  '⇔', '⊂',
    ]

    parms.math_operators.extend(math_operators)

    macros_latex = r"""

        \newcommand{\unimathsetup}[1]{}

    """

    macros_python = [
        Macro(parms, '\\setmathfont', args='AO', repl='')
    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)




#
#   YaLafi module for LaTeX package amsmath
#

from yalafi.defs import Macro, InitModule, EquEnv, SpecialToken

require_packages = []

#
#   Please note:
#   - For all undeclared maths macros, which are not blacklisted in
#     Parameters.math_ignore (yalafi/parameters.py), we assume that
#     a part of a mathematical term or an operator is left.
#   - The spacing macros and \notag are imporant for correct parsing
#     of maths material.
#
def init_module(parser, options, position):
    parms = parser.parms

    parms.math_text_macros.append('\\text')

    macros_latex = r"""

        \newcommand{\eqref}[1]{(0)}
        \newcommand{\medspace}{\:}
        \newcommand{\negmedspace}{}
        \newcommand{\negthickspace}{}
        \newcommand{\negthinspace}{}
        \newcommand{\notag}{}
        \newcommand{\numberwithin}[2]{}
        \newcommand{\thickspace}{\;}
        \newcommand{\thinspace}{\,}

    """

    macros_python = [

        Macro(parms, '\\DeclareMathOperator', args='*AA'),
        Macro(parms, '\\substack', args='A', repl=h_substack),

    ]

    environments = [

        EquEnv(parms, 'align'),
        EquEnv(parms, 'align*'),
        EquEnv(parms, 'alignat', args='A'),
        EquEnv(parms, 'alignat*', args='A'),
        EquEnv(parms, 'equation'),
        EquEnv(parms, 'equation*'),
        EquEnv(parms, 'flalign'),
        EquEnv(parms, 'flalign*'),
        EquEnv(parms, 'gather'),
        EquEnv(parms, 'gather*'),
        EquEnv(parms, 'multiline'),
        EquEnv(parms, 'multiline*'),

    ]

    inject_tokens = []      # this list of tokens is used as replacement
                            # for '\documentclass...' or '\usepackage...'

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                    environments=environments, inject_tokens=inject_tokens)


#   macro \substack: replace \\ tokens with space
#
def h_substack(parser, buf, mac, args, delim, pos):
    def f(tok, lev):
        if tok.txt == r'\\' and lev == 0:
            return SpecialToken(tok.pos, r'\;')
        return tok
    return [f(tok, lev) for tok, lev in parser.iter_token_levels(args[0])]





#
#   YaLafi module for LaTeX package geometry
#

from yalafi.defs import Macro, InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = r"""

        \newcommand{\geometry}[1]{}

    """

    return InitModule(macros_latex=macros_latex)





#
#   YaLafi module for LaTeX package xspace
#

from yalafi import defs
from yalafi.defs import Macro, InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = ""

    macros_python = [

        Macro(parms, '\\xspace', args='', repl=h_xspace),

    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)


#   see http://mirrors.ctan.org/macros/latex/required/tools/xspace.dtx
#   line 243: \def\@xspace@exceptions@tlp{%
#
xspace_excl = ['.', ',', "'", "''", '/', '?', ';', ':', '!', '~',
                '-', '--', '---', ')', '{', '}',
                '\\footnote', '\\footnotemark']


def h_xspace(parser, buf, mac, args, delim, pos):
    tok = buf.cur()
    if tok and tok.txt not in xspace_excl:
        return [defs.SpaceToken(pos, ' ')]
    return []




#   this list is used by yalafi/tex2txt.py
#
load_table = {
    '*': [
        'amsmath',
        'amsthm',
        'babel',
        'biblatex',
        'circuitikz',
        # 'cleveref',
        # cleveref is not loaded by default because it throws
        # a warning if loaded without the poorman option.
        'geometry',
        'glossaries',
        'glossaries-extra',
        'graphicx',
        'hyperref',
        'inputenc',
        'listings',
        'mathtools',
        'pgfplots',
        'tikz',
        'unicode-math',
        'xcolor',
        'xspace',
    ],
    '': [
    ],
}




#
#   YaLafi module for LaTeX package mathtools
#

from yalafi.defs import InitModule

require_packages = ["amsmath"]


def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = r"""

        \newcommand{\mathtoolsset}[1]{}

    """

    macros_python = []

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                      environments=environments)




#
#   YaLafi: \documentclass{scrbook}
#

from yalafi.defs import InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    parser.global_latex_options += options

    macros_latex = r"""

        \newcommand{\KOMAoption}[1]{}
        \newcommand{\KOMAoptions}[1]{}

    """

    return InitModule(macros_latex=macros_latex)




#
#   YaLafi: Yet another LaTeX filter
#   Copyright (C) 2020 Matthias Baumann
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

#
#   default parameters for scanner and parser
#

from .defs import Environ, EquEnv, Macro
from . import handlers as hs
from . import scanner


class Parameters:

    #   pre-defined macros
    #
    def init_macros(self):

        #   definition of macros, LaTeX code
        #
        self.macro_defs_latex = r"""

        \newcommand{\AA}{Å}
        \newcommand{\aa}{å}
        \newcommand{\AE}{Æ}
        \newcommand{\ae}{æ}
        \newcommand{\bibitem}[1]{\item}
        \newcommand{\bibliographystyle}[1]{}
        \newcommand{\footnotemark}[1][]{}
        \newcommand{\hfill}{ }
        \newcommand{\include}[1]{}
        \newcommand{\index}[1]{}
        \newcommand{\input}[1]{}
        \newcommand{\L}{Ł}
        \newcommand{\l}{ł}
        \newcommand{\label}[1]{}
        \newcommand{\LaTeX}{LaTeX}
        \newcommand{\newline}{ }
        \newcommand{\nobreakspace}{~}
        \newcommand{\O}{Ø}
        \newcommand{\o}{ø}
        \newcommand{\OE}{Œ}
        \newcommand{\oe}{œ}
        \newcommand{\pagenumbering}[1]{}
        \newcommand{\pageref}[1]{0}
        \newcommand{\pagestyle}[1]{}
        \newcommand{\par}{

}
        \newcommand{\qquad}{\;}
        \newcommand{\quad}{\;}
        \newcommand{\ref}[1]{0}
        \newcommand{\S}{§}
        \newcommand{\ss}{ß}
        \newcommand{\TeX}{TeX}
        \newcommand{\textasciicircum}{\verb?^?} % \^ is accent
        \newcommand{\textasciitilde}{\verb?~?}  % \~ is accent
        \newcommand{\textbackslash}{\verb?\?}   % \\ is line break
        \newcommand{\thispagestyle}[1]{}
        \newcommand{\vphantom}[1]{}

        """

        #   definition of macros, Python code
        #
        self.macro_defs_python = [

        Macro(self, '\\caption', args='OA', extract='#2'),
        Macro(self, '\\chapter', args='*OA', repl=hs.h_heading),
        Macro(self, '\\cite', args='OA', repl=hs.h_cite),
        Macro(self, '\\documentclass', args='OA',
                            repl=hs.h_load_module(self.class_modules)),
        Macro(self, '\\footnote', args='OA', extract='#2'),
        Macro(self, '\\footnotetext', args='OA', extract='#2'),
        Macro(self, '\\framebox', args='OOA', repl='#3'),
        Macro(self, '\\hphantom', args='A', repl=hs.h_phantom),
        Macro(self, '\\hspace', args='*A', repl=' '),
        Macro(self, '\\newcommand', args='*AOOA', repl=hs.h_newcommand),
        Macro(self, '\\newtheorem', args='AOAO', repl=hs.h_newtheorem),
        Macro(self, '\\part', args='*OA', repl=hs.h_heading),
        Macro(self, '\\phantom', args='A', repl=hs.h_phantom),
        Macro(self, '\\renewcommand', args='*AOOA', repl=hs.h_newcommand),
        Macro(self, '\\section', args='*OA', repl=hs.h_heading),
        Macro(self, '\\subsection', args='*OA', repl=hs.h_heading),
        Macro(self, '\\subsubsection', args='*OA', repl=hs.h_heading),
        Macro(self, '\\title', args='*OA', repl=hs.h_heading),
        Macro(self, '\\usepackage', args='OA',
                            repl=hs.h_load_module(self.package_modules)),
        Macro(self, '\\vspace', args='*A', repl=' '),

        #   \LTadd etc.
        #
        Macro(self, self.macro_filter_add, args='A', repl='#1'),
        Macro(self, self.macro_filter_alter, args='AA', repl='#2'),
        Macro(self, self.macro_filter_skip, args='A', repl=''),
        Macro(self, self.macro_load_defs, args='A', repl=hs.h_load_defs),

        ]

    #   pre-defined environments
    #
    def init_environments(self):

        self.environment_defs = [

        Environ(self, 'figure', args='O', add_pars=False),
        Environ(self, 'minipage', args='A'),
#       Environ(self, 'table', repl='[Tabelle]', remove=True),
        Environ(self, 'table', args='O', add_pars=False),
        Environ(self, 'tabular', args='A', add_pars=False),
        Environ(self, 'thebibliography', args='A', add_pars=True),
        Environ(self, 'verbatim', remove=False, add_pars=True),

        EquEnv(self, 'displaymath'),
        EquEnv(self, 'eqnarray'),
        EquEnv(self, 'eqnarray*'),
        EquEnv(self, 'equation'),

        ]

        #   enumerate: generate 1., 2., 3., ...
        #
        def labs_enumerate(level):
            if level == 0:
                n = 1
                while True:
                    yield str(n) + '.'
                    n += 1
            else:
                c = 'a'
                while True:
                    yield c + '.'
                    c = chr(ord(c) + 1) if c != 'z' else 'a'

        #   itemize: use default labels
        #
        def labs_itemize(level):
            while True:
                yield self.item_default_label[
                            min(level, len(self.item_default_label) - 1)]

        self.environment_defs += [

        Environ(self, 'enumerate', add_pars=False, items=labs_enumerate),
        Environ(self, 'itemize', add_pars=False, items=labs_itemize),

        ]

    #   set language-dependent parser parameters
    #   - settings for 'en' are taken as fall back
    #
    def init_parser_languages(self, lang):
        settings = self.parser_lang_settings = {}

        settings['en'] = ParserLanguageSettings(
            proof_name = 'Proof',
            math_repl_inline = ['B-B-B', 'C-C-C', 'D-D-D',
                                        'E-E-E', 'F-F-F', 'G-G-G'],
            math_repl_inline_vowel = None,
            math_repl_display = ['U-U-U', 'V-V-V', 'W-W-W',
                                        'X-X-X', 'Y-Y-Y', 'Z-Z-Z'],
            math_repl_display_vowel = None,
            math_op_text = {'+': 'plus', '-': 'minus',
                                    '\\cdot': 'times', '\\times': 'times',
                                    '/': 'over',
                                    None: 'equal'},     # default value
            lang_change_repl = ['K-K-K', 'L-L-L', 'M-M-M', 'N-N-N'],
            lang_change_repl_vowel = None,
            short_macros = {}
        )
        settings['de'] = ParserLanguageSettings(
            proof_name = 'Beweis',
            math_repl_inline = ['B-B-B', 'C-C-C', 'D-D-D',
                                        'E-E-E', 'F-F-F', 'G-G-G'],
            math_repl_inline_vowel = None,
            math_repl_display = ['U-U-U', 'V-V-V', 'W-W-W',
                                        'X-X-X', 'Y-Y-Y', 'Z-Z-Z'],
            math_repl_display_vowel = None,
            math_op_text = {'+': 'plus', '-': 'minus',
                                    '\\cdot': 'mal', '\\times': 'mal',
                                    '/': 'durch',
                                    None: 'gleich'},    # default value
            lang_change_repl = ['K-K-K', 'L-L-L', 'M-M-M', 'N-N-N'],
            lang_change_repl_vowel = None,
            short_macros = {
                '"-': '',
                '"=': '-',
                '"`': '\N{DOUBLE LOW-9 QUOTATION MARK}',    # \glqq
                '"\'': '\N{LEFT DOUBLE QUOTATION MARK}',    # \grqq
                '"A': 'Ä',
                '"a': 'ä',
                '"O': 'Ö',
                '"o': 'ö',
                '"U': 'Ü',
                '"u': 'ü',
                '"s': 'ß',
            }
        )
        settings['ru'] = ParserLanguageSettings(
            proof_name = 'Доказательство',
            math_repl_inline = ['Б-Б-Б', 'В-В-В', 'Г-Г-Г',
                                        'Д-Д-Д', 'Е-Е-Е', 'Ж-Ж-Ж'],
            math_repl_inline_vowel = None,
            math_repl_display = ['Ц-Ц-Ц', 'Ч-Ч-Ч', 'Ш-Ш-Ш',
                                        'Ы-Ы-Ы', 'Э-Э-Э', 'Ю-Ю-Ю'],
            math_repl_display_vowel = None,
            math_op_text = {'+': 'плюс', '-': 'минус',
                                    '\\cdot': 'раз', '\\times': 'раз',
                                    '/': 'на',
                                    None: 'равно'},     # default value
            lang_change_repl = ['К-К-К', 'Л-Л-Л', 'М-М-М', 'Н-Н-Н'],
            lang_change_repl_vowel = None,
            short_macros = {}
        )

        self.parser_lang_stack = [(settings[self.check_parser_lang(lang)],
                                        lang)]
        self.lang_context = self.parser_lang_stack[-1][0]

    #   set misc collections
    #
    def init_collections(self):

        #   add dot to heading unless last heading char in ...
        #   (turn off: set to [])
        #
        self.heading_punct = ['!', '?']

        #   labels for \item without [...] (may depend on nesting level)
        #
        self.item_default_label = ['']

        #   \item with [...] label: if text before ends with something
        #   from here, then append it to label
        #
        self.item_punctuation = ['.', ':', ',', ';', '!', '?']

        #   mark for parsing problem, should raise message from proofreader
        #
        self.mark_latex_error = 'LATEXXXERROR'

        #   True: include error description
        #
        self.mark_latex_error_verbose = False

        #   names of special macros
        #
        self.macro_filter_add = '\\LTadd'
        self.macro_filter_alter = '\\LTalter'
        self.macro_filter_skip = '\\LTskip'
        self.macro_load_defs = '\\LTinput'

        #   ignore re-definitions in LaTeX text for these macros:
        #
        self.newcommand_ignore = [
            self.macro_filter_add,
            self.macro_filter_alter,
            self.macro_filter_skip,
            self.macro_load_defs,
        ]

        #   special LaTeX comments to skip input text
        #
        self.comment_skip_begin = '%%% LT-SKIP-BEGIN'
        self.comment_skip_end = '%%% LT-SKIP-END'

        #   module directories
        #
        self.class_modules = 'yalafi.documentclasses'
        self.package_modules = 'yalafi.packages'

        #   accent macros
        #
        self.accent_macros = {

            "\\'": ['ACUTE', 'ACCENT'],
            '\\`': ['GRAVE', 'ACCENT'],
            '\\^': ['CIRCUMFLEX', 'ACCENT'],
            '\\v': ['CARON'],
            '\\~': ['TILDE'],
            '\\"': ['DIAERESIS'],
            '\\r': ['RING ABOVE'],
            '\\=': ['MACRON'],
            '\\b': ['LINE BELOW'],
            '\\u': ['BREVE'],
            '\\H': ['DOUBLE ACUTE', 'ACCENT'],
            '\\.': ['DOT ABOVE'],
            '\\d': ['DOT BELOW'],
            '\\c': ['CEDILLA'],
            '\\k': ['OGONEK'],

        }

        #   "special" tokens: expanded to key value
        #
        self.special_tokens = {

            '{': '{',
            '}': '}',

            '$$': '$$',
            '$': '$',
            '#': '#',
            '&': ' ',       # expand to space
            '_': '_',
            '^': '^',

            '\\(': '\\(',
            '\\)': '\\)',
            '\\[': '\\[',
            '\\]': '\\]',
            '\\\\': ' ',    # expand to space

            '~': '\N{NO-BREAK SPACE}',
            '``': '\N{LEFT DOUBLE QUOTATION MARK}',
            "''": '\N{RIGHT DOUBLE QUOTATION MARK}',
            '--': '\N{EN DASH}',
            '---': '\N{EM DASH}',

            '\\ ': ' ',
            '\\\t': ' ',
            '\\\n': ' ',
            '\\,': '\N{NARROW NO-BREAK SPACE}',
            '\\:': ' ',
            '\\;': ' ',
            '\\!': '',

            '\\{': '{',
            '\\}': '}',
            '\\$': '$',
            '\\#': '#',
            '\\&': '&',
            '\\_': '_',
            '\\%': '%',

        }

    #   set math collections
    #
    def init_math_collections(self):

        #   simple replacement for displayed equations?
        #
        self.math_displayed_simple = False

        #   things to be ignored in math mode
        #   - some entries are redundant, if macros are known from text mode
        #     and expand to 'nothing'
        #
        self.math_ignore = [

            '{',
            '}',
            '\\!',
            '\\label',
            '\\mathrlap',
            '\\nonumber',

        ]

        #   things that generate space even in math mode
        #
        self.math_space = [

            '~',
            '\\ ',
            '\\\t',
            '\\\n',
            '\\:',
            '\\,',
            '\\;',

        ]

        #   if these operators appear first in a part (only math tokens)
        #   then one of the replacements in self.math_op_text is inserted
        #
        self.math_operators = [

            '+', '-', '\\cdot', '\\times', '/',
            '=', '\\eq', '\\ne', '\\neq',
            '<', '>', '\\le', '\\leq', '\\ge', '\\geq',
            ':', ':=', '\\to', '\\cap', '\\cup',
            '\\Rightarrow', '\\Leftarrow', '\\Leftrightarrow',
            '\\subset', '\\subseteq', '\\supset', '\\supseteq',
            '\\stackrel',

        ]

        #   macros whose argument is treated in text mode
        #
        self.math_text_macros = [

            '\\mbox',

        ]

        #   math environment for $$ and \[
        #
        self.math_default_env = 'displaymath'

        #   if a math section ends with a character from here, it is
        #   appended to the replacement from self_repl_xxx
        #
        self.math_punctuation = ['.', ',', ';', ':']

    #   determine whether a character may be part of macro name
    #
    def macro_character(self, c):
        return c >= 'a' and c <= 'z' or c >= 'A' and c <= 'Z' or c == '@'

    #   transform given lang into valid key for dictionary
    #   self.parser_lang_settings
    #
    def check_parser_lang(self, lang):
        lang = lang[:2].lower()
        return lang if lang in self.parser_lang_settings else 'en'

    #   switch current parser language settings to new language
    #
    def change_parser_lang(self, tok):
        if tok.back:
            if len(self.parser_lang_stack) > 1:
                self.parser_lang_stack.pop()
        else:
            if tok.hard:
                self.parser_lang_stack[-1] = (
                self.parser_lang_settings[self.check_parser_lang(tok.lang)],
                        tok.lang)
            else:
                self.parser_lang_stack.append(
                (self.parser_lang_settings[self.check_parser_lang(tok.lang)],
                        tok.lang))
        self.lang_context = self.parser_lang_stack[-1][0]

    def lang_context_lang(self):
        return self.parser_lang_stack[-1][1]

    #   deactivate special macros and magic comments
    #
    def no_specials(self):
        self.comment_skip_begin = 'x'
        self.comment_skip_end = 'x'
        self.macro_defs_python += [
            Macro(self, self.macro_filter_add, args='A', repl=''),
            Macro(self, self.macro_filter_alter, args='AA', repl='#1'),
            Macro(self, self.macro_filter_skip, args='A', repl='#1'),
        ]

    def __init__(self, language='en'):
        self.init_collections()
        self.init_math_collections()
        self.multi_language = False
        self.ml_continue_thresh = 3
        self.init_parser_languages(language)
        self.scanner = scanner.Scanner(self)
        self.init_macros()
        self.init_environments()


class ParserLanguageSettings:
    def __init__(self, proof_name,
                        math_repl_inline, math_repl_inline_vowel,
                        math_repl_display, math_repl_display_vowel,
                        math_op_text,
                        lang_change_repl, lang_change_repl_vowel,
                        short_macros):
        self.proof_name = proof_name
        self.math_repl_inline = math_repl_inline
        if math_repl_inline_vowel is None:
            self.math_repl_inline_vowel = math_repl_inline
        else:
            self.math_repl_inline_vowel = math_repl_inline_vowel
        self.math_repl_display = math_repl_display
        if math_repl_display_vowel is None:
            self.math_repl_display_vowel = math_repl_display
        else:
            self.math_repl_display_vowel = math_repl_display_vowel
        self.math_op_text = math_op_text
        self.lang_change_repl = lang_change_repl
        if lang_change_repl_vowel is None:
            self.lang_change_repl_vowel = lang_change_repl
        else:
            self.lang_change_repl_vowel = lang_change_repl_vowel
        self.short_macros = short_macros
        self.active_chars = set(k[0] for k in self.short_macros.keys())





#
#   YaLafi module for LaTeX package glossaries-extra
#
#   - In order to let LaTeX generate the .glsdefs file, say
#       \usepackage[docdef=true]{glossaries-extra}
#   - Please note the comments at the beginning of glossaries.py.
#   - You can keep all definitions of glossary entries in the preamble,
#     if you say
#       \usepackage[docdef=atom]{glossaries-extra}
#

from yalafi.defs import Macro, InitModule

require_packages = ['glossaries']

def init_module(parser, options, position):
    parms = parser.parms

    macros_latex = ''

    macros_python = [

        Macro(parms, '\\newabbreviation', args='AAA',
                                            repl=r'\newacronym{#1}{#2}{#3}'),

    ]

    environments = []

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                        environments=environments)





#
#   YaLafi module for LaTeX package cleveref
#   Contributed by @torik42 (at GitHub) in pull request #171
#

from yalafi.defs import InitModule, Macro, VoidToken
from yalafi import utils
import re


require_packages = []

macro_read_sed = '\\YYCleverefInput'

# Regular Expressions used to read sed file:
re_ref = re.compile(r'''
s/\\
(                   # group 1: '\cref' or '\Cref'
    \\cref|\\Cref
)
(?:\\)?             # non capturing group
(\*?)               # group 2: '*' if present, '' else
\{
    ([^\}\{]+)      # group 3: the label of the reference
\}
/
    (.*)            # group 4: the replacement string
/g
''', re.VERBOSE)
re_ref_range = re.compile(r'''
s/\\
(                   # group 1: '\cref' or '\Cref'
    \\crefrange|\\Crefrange
)
(?:\\)?             # non capturing group
(\*?)               # group 2: '*' if present, '' else
\{
    ([^\}\{]+)      # group 3: the label of the first reference
\}
\{
    ([^\}\{]+)      # group 4: the label of the second reference
\}
/
    (.*)            # group 5: the replacement string
/g
''', re.VERBOSE)
re_command = re.compile(r'''
s/\\
(                   # group 1: command
    \\
    (               # group 2: capturing cC
        \[cC\]
    )?
    [a-zA-Z@]+
)
(                   # group 3: possible arguments
    (?:\{\.\*\})+
)?
\s*                 # possible whitespaces inbetween
/
    (.*)            # group 4: replacement
/g
''', re.VERBOSE)
re_cC = re.compile(r'(\[cC\])')
re_escaped_dot = re.compile(r'\\\.')
re_escaped_backslash = re.compile(r'\\\\')
re_escaped_star = re.compile(r'\\\*')


def re_remove_escaped_symbols(string):
    r"""
    Replaces certain strings that are escaped in sed file:
    '\\.' -> '.'
    '\\\\' -> '\\'
    '\*' -> ''
    """
    string = re_escaped_dot.sub('.', string)
    string = re_escaped_star.sub('', string)
    string = re_escaped_backslash.sub(r'\\', string)
    return string


# Error messages
msg_poorman_option = f'''To use cleveref with YaLafi, you need to use
*** the 'poorman' option and {macro_read_sed}
*** to load the sed file.
'''
msg_sed_not_loaded = f'''To use cleveref with YaLafi, you should use
*** {macro_read_sed} to load the sed file, e.g.
*** '{macro_read_sed}{{main.sed}}' if your LaTeX
*** document is called 'main.tex'.
'''

msg_cref_undefined = r'''No replacement for {:}{{{:}}} known.
*** Run LaTeX again to build a new sed file.
'''
msg_crefrange_undefined = r'''No replacement for {:}{{{:}}}{{{:}}} known.
*** Run LaTeX again to build a new sed file.
'''


def init_module(parser, options, position):
    parms = parser.parms
    parms.newcommand_ignore.append(macro_read_sed)

    macros_latex = r'''
        \newcommand{\crefname}[3]{}
        \newcommand{\Crefname}[3]{}
    '''

    macros_python = [

        Macro(parms, macro_read_sed, args='A', repl=h_read_sed),
        Macro(parms, '\\label', args='OA', repl=''),

        # Define functions which warn the User, whenever cleveref
        # is used without invoking \YYCleverefInput. These will be overwritten
        # by invoking \YYCleverefInput.
        Macro(parms, '\\cref', args='*A', repl=h_cref_warning),
        Macro(parms, '\\Cref', args='*A', repl=h_cref_warning),
        Macro(parms, '\\crefrange', args='*AA', repl=h_cref_warning),
        Macro(parms, '\\Crefrange', args='*AA', repl=h_cref_warning),

    ]

    environments = []

    # Warn the user, whenever cleveref is used
    # without the poorman option:
    inject_tokens = []
    if not is_poorman_used(options):
        inject_tokens = utils.latex_error(msg_poorman_option,
                                          position, parser.latex, parms)

    return InitModule(macros_latex=macros_latex, macros_python=macros_python,
                      environments=environments, inject_tokens=inject_tokens)


def h_read_sed(parser, buf, mac, args, delim, pos):
    if not parser.read_macros:
        return []

    # Read sed file into sed:
    file = parser.get_text_expanded(args[0])
    ok, sed = parser.read_macros(file)

    # Throw LaTeX error if the file could not be loaded:
    if not ok:
        return utils.latex_error('could not read file ' + repr(file),
                                 pos, parser.latex, parser.parms)

    refs = {'\\cref': {'':{}, '*': {}},
            '\\Cref': {'':{}, '*': {}},
            '\\crefrange': {'':{}, '*': {}},
            '\\Crefrange': {'':{}, '*': {}}}

    for rep in sed.split('\n'):
        # only consider non-empty lines:
        if rep == '':
            continue

        # Match \cref,\cref*,\Cref and \Cref* and save the replacement string:
        m = re_ref.match(rep)
        if m:
            refs[m.group(1)][m.group(2)][m.group(3)] \
                = re_remove_escaped_symbols(m.group(4))
            continue

        # Match \crefrange, \crefrange*, \Crefrange and \Crefrange* and
        # save the replacement string:
        m = re_ref_range.match(rep)
        if m:
            refs[m.group(1)][m.group(2)][(m.group(3), m.group(4))] \
                = re_remove_escaped_symbols(m.group(5))

        # Match any other command and create Macro objects for them.
        # See definition of re_command for more details:
        m = re_command.match(rep)
        if m:
            args = 'A'*int((m.end(3)-m.start(3))/4)
            string = re_remove_escaped_symbols(m.group(4))
            if m.group(2):
                name = re_cC.sub('c', m.group(1))
                parser.the_macros[name] = Macro(parser.parms,
                                                name, args=args, repl=string)
                name = re_cC.sub('C', m.group(1))
                parser.the_macros[name] = Macro(parser.parms,
                                                name, args=args, repl=string)
            else:
                name = m.group(1)
                parser.the_macros[name] = Macro(parser.parms,
                                                name, args=args, repl=string)

    # Make the \cref, …, \Crefrange* Macro objects:
    for ref in ['\\cref','\\Cref']:
        parser.the_macros[ref] = Macro(parser.parms,
                                       ref, args='*A',
                                       repl=h_make_cref(refs[ref]))
    for ref in ['\\crefrange','\\Crefrange']:
        parser.the_macros[ref] = Macro(parser.parms,
                                       ref, args='*AA',
                                       repl=h_make_crefrange(refs[ref]))

    # \YYCleverefInput should not produce any output:
    return []


def h_make_cref(cref):
    def f(parser, buf, mac, args, delim, pos):
        star = parser.get_text_direct(args[0])
        rep = parser.get_text_direct(args[1])
        if rep in cref[star]:
            toks = parser.parms.scanner.scan(cref[star][rep])
            for t in toks:
                t.pos = pos
            return toks
        return utils.latex_error(msg_cref_undefined.format(mac.name,rep),
                                 pos, parser.latex, parser.parms)
    return f


def h_make_crefrange(cref):
    def f(parser, buf, mac, args, delim, pos):
        star = parser.get_text_direct(args[0])
        rep = (parser.get_text_direct(args[1]), parser.get_text_direct(args[2]))
        if rep in cref[star]:
            toks = parser.parms.scanner.scan(cref[star][rep])
            for t in toks:
                t.pos = pos
            return toks
        return utils.latex_error(msg_crefrange_undefined.format(mac.name,*rep),
                                 pos, parser.latex, parser.parms)
    return f


def h_cref_warning(parser, buf, mac, args, delim, pos):
    return utils.latex_error(msg_sed_not_loaded,
                             pos, parser.latex, parser.parms)


def is_poorman_used(options):
    for opt in reversed(options):
        if 'poorman' in opt:
            return True
    return False




#
#   YaLafi: \documentclass{book}
#

from yalafi.defs import InitModule

require_packages = []

def init_module(parser, options, position):
    parms = parser.parms

    parser.global_latex_options += options

    return InitModule()

